"""
Defines classes which represent bonds.

"""


class Bond:
    """
    Represents an atomic bond.

    Attributes
    ----------
    atom1 : :class:`.Atom`
        The first atom in the bond.

    atom2 : :class:`.Atom`
        The second atom in the bond.

    order : :class:`int`
        The bond order.

    periodicity : :class:`tuple` of :class:`int`
        The directions across which the bond is periodic. For example,
        ``(1, 0, -1)`` means that when going from
        :attr:`~.Bond.atom1` to :attr:`~.Bond.atom2` the bond is
        periodic across the x axis in the positive direction, is not
        periodic across the y axis and is periodic across the z axis
        in the negative direction.

    Examples
    --------
    *Adding additional attributes.*

    Each bond can be given additional attributes. For example

    .. code-block:: python

        import stk

        bond1 = stk.Bond(stk.H(0), stk.H(1), 1)
        bond1.attr1 = 12.2
        bond1.attr2 = 'something'

    A :class:`.Bond` can also initialized with additional attributes
    directly

    .. code-block:: python

        bond2 = stk.Bond(
            atom1=stk.C(2),
            atom2=stk.C(3),
            order=1,
            attr1=123,
            attr2='hi'
        )

        bond2.attr1  # Holds 123.
        bond2.attr2  # Holds 'hi'.

    *Printing*

    To print a brief summary of a bond you can run

    .. code-block:: python

        # Prints Bond(C(2), C(3), 1).
        print(bond2)

    To print a complete description of a bond, including additional
    attributes, you can run

    .. code-block:: python

        # Prints Bond(C(2), C(3), 1, attr1=123, attr2='hi')
        print(repr(bond2))

    If the atoms have additional attributes, they will be printed too

    .. code-block:: python

        # Add a custom attribute to an atom.
        bond2.atom1.alpha = 1

        # Prints Bond(C(2, alpha=1), C(3), 1, attr1=123, attr2='hi')
        print(repr(bond2))

    If private attributes were added to a bond, they will not be
    printed

    .. code-block:: python

        bond2._attr3 = 'private'
        # Prints Bond(C(2, alpha=1), C(3), 1, attr1=123, attr2='hi')
        print(repr(bond2))

    """

    def __init__(
        self,
        atom1,
        atom2,
        order,
        periodicity=(0, 0, 0),
        **kwargs
    ):
        """
        Initialize a :class:`Bond`.

        Attributes
        ----------
        atom1 : :class:`.Atom`
            The first atom in the bond.

        atom2 : :class:`.Atom`
            The second atom in the bond.

        order : :class:`int`
            The bond order.

        periodicity : :class:`tuple` of :class:`int`, optional
            The directions across which the bond is periodic. For
            example, ``(1, 0, -1)`` means that when going from
            :attr:`~.Bond.atom1` to :attr:`~.Bond.atom2` the bond is
            periodic across the x axis in the positive direction, is
            not periodic across the y axis and is periodic across the z
            axis in the negative direction.

        **kwargs : :class:`object`, optional
            Additional attributes to be added to the bond.

        """

        self.atom1 = atom1
        self.atom2 = atom2
        self.order = order
        self.periodicity = periodicity
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def clone(self, atom_map=None):
        """
        Return a clone.

        Private attributes are not passed to the clone.

        Parameters
        ----------
        atom_map : :class:`dict`, optional
            If the clone should hold different :class:`.Atom`
            instances, then a :class:`dict` should be provided, which
            maps atoms in the current :class:`.Bond` to the
            atoms which should be used in the clone. Only atoms which
            need to be remapped need to be present in the `atom_map`.

        Returns
        -------
        :class:`.Bond`
            The clone.

        Examples
        --------
        .. code-block:: python

            import stk

            c0 = stk.C(0)
            c1 = stk.C(1)
            bond = stk.Bond(c0, c1, 1, custom_attr=12, _private_attr=1)

            # bond_clone holds c0 and c1 in its atom1 and atom2
            # attributes, respectively. It also has a custom_attr
            # with a value of 12 but it does not have a _private_attr
            # attribute.
            bond_clone = bond.clone()

        It is possible to make sure that the clone holds different
        atoms

        .. code-block:: python

            li2 = stk.Li(2)
            n3 = stk.N(3)

            # clone2 is also a clone, except that it holds
            # li2 in the atom2 attribute. Its atom1 attribute still
            # holds c0.
            clone2 = bond.clone(atom_map={
                c1: li2
            })

            # clone3 is also a clone, except that it holds n3 and
            # li2 in its atom1 and atom2 attributes, respectively.
            clone3 = bond.clone(atom_map={
                c0: n3,
                c1: li2
            })

        """

        if atom_map is None:
            atom_map = {}

        obj = self.__class__.__new__(self.__class__)
        for attr, val in vars(self).items():
            if not attr.startswith('_'):
                setattr(obj, attr, val)
        obj.atom1 = atom_map.get(obj.atom1, obj.atom1)
        obj.atom2 = atom_map.get(obj.atom2, obj.atom2)
        return obj

    def is_periodic(self):
        """
        Return ``True`` if the bond is periodic.

        Returns
        -------
        :class:`bool`
            ``True`` if the bond is periodic.

        """

        return any(direction != 0 for direction in self.periodicity)

    def __repr__(self):
        if isinstance(self.order, float) and self.order.is_integer():
            self.order = int(self.order)

        periodicity = (
            '' if not self.is_periodic()
            else f', periodicity={self.periodicity}'
        )

        mandatory = {'atom1', 'atom2', 'order', 'periodicity'}
        attrs = ', '.join(
            f'{attr}={val!r}' for attr, val in vars(self).items()
            if attr not in mandatory and not attr.startswith('_')
        )
        cls_name = self.__class__.__name__
        return (
            f'{cls_name}({self.atom1!r}, {self.atom2!r}, '
            f'{self.order}{periodicity}{", " if attrs else ""}{attrs})'
        )

    def __str__(self):
        cls_name = self.__class__.__name__
        if isinstance(self.order, float) and self.order.is_integer():
            self.order = int(self.order)

        periodicity = (
            '' if not self.is_periodic()
            else f', {self.periodicity}'
        )
        return (
            f'{cls_name}({self.atom1}, {self.atom2}, '
            f'{self.order}{periodicity})'
        )



"""
Macrocycle
==========

#. :class:`.Macrocycle`

"""

import numpy as np
import logging
from scipy.spatial.distance import euclidean

from .topology_graph import TopologyGraph, VertexData, Vertex, EdgeData


logger = logging.getLogger(__name__)


class _CycleVertexData(VertexData):
    """
    Holds data for a :class:`.CycleVertex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    flip : :class:`bool`
        If ``True`` any building block placed by the vertex will
        have its orientation along the chain flipped.

    angle : :class:`float`
        The angle along the macrocycle at which the vertex is
        found.

    """

    def __init__(self, x, y, z, flip, angle):
        """
        Initialize a :class:`._CycleVertexData`.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

    flip : :class:`bool`
        If ``True`` any building block placed by the vertex will
        have its orientation along the chain flipped.

        angle : :class:`float`
            The angle along the macrocycle at which the vertex is
            found.

        """

        self.flip = flip
        self.angle = angle
        super().__init__(x, y, z)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone.flip = self.flip
        clone.angle = self.angle
        return clone

    def get_vertex(self):
        return _CycleVertex(self)


class _CycleVertex(Vertex):
    """
    Represents a vertex in the middle of a linear polymer chain.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        self._flip = data.flip
        self._angle = data.angle
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            If ``True`` the :attr:`edges` attribute of the clone will
            be empty.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone._flip = self._flip
        clone._angle = self._angle
        return clone

    def place_building_block(self, building_block, vertices, edges):
        if len(building_block.func_groups) > 2:
            logger.warning(
                'You are placing a building block which has more than '
                'two functional groups along the backbone of '
                'a Macrocycle topology. You can remove extra '
                'functional groups from the func_groups attribute to '
                'remove this message.'
            )

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids(fg_ids=(0, 1))
        )
        bonder_vector = next(
            building_block.get_bonder_direction_vectors(
                fg_ids=(0, 1)
            )
        )[-1]
        building_block.apply_rotation_between_vectors(
            start=bonder_vector,
            target=[-1 if self._flip else 1, 0, 0],
            origin=self._position
        )
        building_block.apply_rotation_about_axis(
            angle=self._angle-(np.pi/2),
            axis=np.array([0, 0, 1]),
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return {
            fg_id: edge_id for fg_id, edge_id in enumerate(sorted(
                self._edge_ids,
                key=lambda edge_id: self._fg0_distance(
                    building_block=building_block,
                    edge_id=edge_id,
                    edges=edges
                )
            ))
        }

    def _fg0_distance(self, building_block, edge_id, edges):
        fg_position = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        return euclidean(edges[edge_id].get_position(), fg_position)

    def __str__(self):
        return (
            f'Vertex(id={self.id}, '
            f'position={self._position.tolist()}, '
            f'flip={self._flip}, '
            f'angle={self._angle})'
        )


class Macrocycle(TopologyGraph):
    """
    Represents a macrocycle topology graph.

    The macrocycle can be represented as a linear polymer with the two
    end groups bonded to close the loop.

    Attributes
    ----------
    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    .. code-block:: python

        import stk

        macrocycle = stk.ConstructedMolecule(
            building_blocks=[
                stk.BuildingBlock('NCCN', ['amine']),
                stk.BuildingBlock('O=CCC=O', ['aldehyde'])
            ],
            topology_graph=stk.macrocycle.Macrocycle('AB', 5)
        )

    The repeating unit can also be specified through the indices of
    the building blocks

    .. code-block:: python

        bb1 = stk.BuildingBlock('BrCCBr', ['bromine'])
        bb2 = stk.BuildingBlock('BrCNCBr', ['bromine'])
        bb3 = stk.BuildingBlock('BrCNNCBr', ['bromine'])

        # c1 and c2 are different ways to write the same thing.
        c1 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=stk.macrocycle.Macrocycle('ACB', 3)
        )
        c2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=stk.macrocycle.Macrocycle((0, 2, 1), 3)
        )

    :class:`.Macrocycle` shares many parameters with :class:`.Linear`,
    and the examples described there are also valid for this class.
    Be sure to read them.

    """

    def __init__(
        self,
        repeating_unit,
        num_repeating_units,
        orientations=None,
        random_seed=None,
        num_processes=1
    ):
        """
        Initialize a :class:`Macrocycle` instance.

        Parameters
        ----------
        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            A string specifying the repeating unit of the macrocycle.
            For example, ``'AB'`` or ``'ABB'``. The first building
            block passed to `building_blocks` is ``'A'`` and so on.

            The repeating unit can also be specified by the indices of
            `building_blocks`, for example ``'ABB'`` can be
            written as ``(0, 1, 1)``.

        num_repeating_units : :class:`int`
            The number of repeating units which are used to make the
            macrocycle.

        orientations : :class:`tuple` of :class:`float`, optional
            For each character in the repeating unit, a value
            between ``0`` and ``1`` (both inclusive) must be given in
            a :class:`tuple`. It indicates the probability that each
            monomer will have its orientation along the chain flipped.
            If ``0`` then the monomer is guaranteed not to flip. If
            ``1`` it is guaranteed to flip. This allows the user to
            create head-to-head or head-to-tail chains, as well as
            chain with a preference for head-to-head or head-to-tail if
            a number between ``0`` and ``1`` is chosen. If ``None``
            then ``0`` is picked in every case.

            It is also possible to supply an orientation for every
            vertex in the final topology graph. In this case, the
            length of `orientations` must be equal to
            ``len(repeating_unit)*num_repeating_units``.

        random_seed : :class:`int`, optional
            The random seed to use when choosing random orientations.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        """

        if orientations is None:
            orientations = tuple(
                0. for i in range(len(repeating_unit))
            )

        if len(orientations) == len(repeating_unit):
            orientations = orientations*num_repeating_units

        chain_length = len(repeating_unit)*num_repeating_units
        if len(orientations) != chain_length:
            raise ValueError(
                'The length of orientations must match either '
                'the length of repeating_unit or the '
                'total number of vertices.'
            )

        generator = np.random.RandomState(random_seed)

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        # Each monomer in the macrocycle is separated by angle_diff.
        angle_diff = (2*np.pi)/chain_length
        vertex_data = []
        edge_data = []
        choices = [True, False]
        for i, p in enumerate(orientations):
            theta = i*angle_diff
            v = _CycleVertexData(
                x=np.cos(theta),
                y=np.sin(theta),
                z=0,
                flip=generator.choice(choices, p=[p, 1-p]),
                angle=theta
            )
            vertex_data.append(v)

            if i > 0:
                edge_data.append(
                    EdgeData(vertex_data[i-1], vertex_data[i])
                )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(int(v.flip) for v in vertex_data)

        edge_data.append(EdgeData(vertex_data[0], vertex_data[-1]))
        super().__init__(
            vertex_data=tuple(vertex_data),
            edge_data=tuple(edge_data),
            construction_stages=(),
            num_processes=num_processes
        )

    @staticmethod
    def _normalize_repeating_unit(repeating_unit):
        if isinstance(repeating_unit, tuple):
            return repeating_unit
        base = ord('A')
        return tuple(ord(letter)-base for letter in repeating_unit)

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        """

        polymer = self._repeating_unit*self._num_repeating_units
        building_block_vertices = {}
        for bb_index, vertex in zip(polymer, self.vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        length = len(self._repeating_unit)*self._num_repeating_units
        return length*0.25*max(
            bb.get_maximum_diameter()
            for bb in mol.building_block_vertices
        )

    def __repr__(self):
        return (
            f'macrocycle.Macrocycle({self._repeating_unit!r}, '
            f'{self._num_repeating_units!r}, '
            f'{self._orientations!r})'
        )



"""
Functional Groups
=================

The purpose of functional groups is to identify which atoms of a
building block are modified during construction. See
:class:`.FunctionalGroup` for more details.

See the documentation of :class:`.Reactor` to see how reactions between
functional groups are performed.

.. _`adding functional groups`:

Adding Functional Groups
------------------------

During initialization of a :class:`.BuildingBlock` instance, the
names of :class:`.FGType` instances are supplied. These are used to
create the :class:`.FunctionalGroup` instances
found in :attr:`.BuildingBlock.func_groups` and therefore the ones
used in the construction of :class:`.ConstructedMolecule` instances.
If you want to be able to use a new :class:`.FGType` with building
blocks in this way, simply add a new :class:`.FGType` instance to
:data:`fg_types`. This can be done in this file or externally.
For example, if you want make ``stk`` use a new :class:`.FGType`
without touching the source code you can just do

.. code-block:: python

    import stk

    new_fg_type = FGType(
        name='my_fg_type',
        func_group_smarts='[N]([H])[H]',
        bonder_smarts=['[$([N]([H])[H])]'],
        deleter_smarts=['[$([H][N][H])]']*2
    )
    stk.fg_types[new_fg_type.name] = new_fg_type

    # You can now use the new_fg_type with BuildingBlocks.
    bb = stk.BuildingBlock('NCCCN', ['my_fg_type'])

Note that when adding SMARTS, if you want to make a SMARTS that targets
an atom in an environment, for example, a bromine connected to a
carbon::

    [$([Br][C])]

The atom you are targeting needs to be written first. The above SMARTS
works but::

    [$([C][Br])]

does not.

"""

import rdkit.Chem.AllChem as rdkit
from collections import Counter

from ..utilities import flatten


class FGType:
    """
    Creates :class:`.FunctionalGroup` instances.

    Attributes
    ----------
    name : :class:`str`
        A name for the :class:`.FGType`. Helps with identification.

    """

    def __init__(
        self,
        name,
        func_group_smarts,
        bonder_smarts,
        deleter_smarts
    ):
        """
        Initialize a :class:`FGType` instance.

        Parameters
        ----------
        name : :class:`str`
            A name for the :class:`.FGType`. Helps with identification.

        func_group_smarts : :class:`str`
            A SMARTS string which matches all atoms in a functional
            group.

        bonder_smarts : :class:`list` of :class:`str`
            A :class:`list` of SMARTS strings, each of which matches a
            single atom in a functional group. The matched atom is
            added to :attr:`.FunctionalGroup.bonders`. A SMARTS
            string needs to be repeated if a single functional group
            has multiple equivalent atoms, each of which appears in
            the same :class:`.FunctionalGroup`.

        deleter_smarts : :class:`list` of :class:`str`
            A :class:`list` of SMARTS strings, each of which matches a
            single atom in a functional group. The matched atom is
            added to :attr:`.FunctionalGroup.deleters`. A SMARTS
            string needs to be repeated if a single functional group
            has multiple equivalent atoms, each of which appears in
            the same :class:`.FunctionalGroup`.

        """

        self.name = name
        self._func_group = rdkit.MolFromSmarts(func_group_smarts)
        self._bonders = [
            (rdkit.MolFromSmarts(smarts), count)
            for smarts, count in Counter(bonder_smarts).items()
        ]
        self._deleters = [
            (rdkit.MolFromSmarts(smarts), count)
            for smarts, count in Counter(deleter_smarts).items()
        ]

    def get_functional_groups(self, mol):
        """
        Yield the functional groups in `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule which is to have functional groups
            identified.

        Yields
        ------
        :class:`.FunctionalGroup`
            A :class:`.FunctionalGroup` for every matched
            functional group in the `mol`.

        Examples
        --------
        .. code-block:: python

            import stk

            mol = stk.BuildingBlock('NCCN')
            amine = stk.FGType(
                func_group_smarts='[N]([H])[H]',
                bonder_smarts=['[$([N]([H])[H])]'],
                deleter_smarts=['[$([H][N][H])]']*2
            )

            # Make mol use amine functional groups during construction
            # of ConstructedMolecule.
            mol.func_groups = tuple(amine.get_functional_groups(mol))

        """

        rdkit_mol = mol.to_rdkit_mol()
        rdkit.SanitizeMol(rdkit_mol)

        func_groups = rdkit_mol.GetSubstructMatches(self._func_group)

        # All the bonder atoms, grouped by fg.
        bonders = [[] for i in range(len(func_groups))]

        for bonder, count in self._bonders:
            matches = set(flatten(
                rdkit_mol.GetSubstructMatches(bonder)
            ))

            matched_bonders = [
                [aid for aid in fg if aid in matches]
                for fg in func_groups
            ]

            for fg_id, fg in enumerate(func_groups):
                bonders[fg_id].extend(matched_bonders[fg_id][:count])

        # All the deleter atoms, grouped by fg.
        deleters = [[] for i in range(len(func_groups))]
        for deleter, count in self._deleters:
            matches = set(flatten(
                rdkit_mol.GetSubstructMatches(deleter)
            ))

            matched_deleters = [
                [aid for aid in fg if aid in matches]
                for fg in func_groups
            ]

            for fg_id, fg in enumerate(func_groups):
                deleters[fg_id].extend(matched_deleters[fg_id][:count])

        for atom_ids in zip(func_groups, bonders, deleters):
            fg, fg_bonders, fg_deleters = atom_ids
            yield FunctionalGroup(
                atoms=tuple(mol.atoms[id_] for id_ in fg),
                bonders=tuple(mol.atoms[id_] for id_ in fg_bonders),
                deleters=tuple(mol.atoms[id_] for id_ in fg_deleters),
                fg_type=self
            )

    def __repr__(self):
        func_group_smarts = rdkit.MolToSmarts(self._func_group)
        bonder_smarts = [
            rdkit.MolToSmarts(mol) for mol, _ in self._bonders
        ]
        deleter_smarts = [
            rdkit.MolToSmarts(mol) for mol, _ in self._deleters
        ]
        return (
            f'FGType(\n'
            f'    name={self.name!r}\n'
            f'    func_group_smarts={func_group_smarts!r},\n'
            f'    bonder_smarts={bonder_smarts!r},\n'
            f'    deleter_smarts={deleter_smarts!r}\n'
            ')'
        )

    def __str__(self):
        return f'FGType({self.name!r})'


class FunctionalGroup:
    """
    Represents a functional group in a molecule.

    Instances of this class should only be made via
    :meth:`.FGType.get_functional_groups`.

    Attributes
    ----------
    atoms : :class:`tuple` of :class:`.Atom`
        The atoms in the functional group.

    bonders : :class:`tuple` of :class:`.Atom`
        The bonder atoms in the functional group. These are atoms which
        have bonds added during the construction of a
        :class:`.ConstructedMolecule`.

    deleters : :class:`tuple` of :class:`.Atom`
        The deleter atoms in the functional group. These are atoms
        which are deleted during construction of a
        :class:`.ConstructedMolecule`.

    fg_type : :class:`.FGType`
        The :class:`.FGType` instance which created the
        :class:`.FunctionalGroup`.

    """

    def __init__(self, atoms, bonders, deleters, fg_type):
        """
        Initialize a :class:`.FunctionalGroup`.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms in the functional group.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms in the functional group.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms in the functional group.

        fg_type : :class:`.FGType`
            The :class:`.FGType` instance which created the
            :class:`.FunctionalGroup`.

        """

        self.atoms = atoms
        self.bonders = bonders
        self.deleters = deleters
        self.fg_type = fg_type

    def clone(self, atom_map=None):
        """
        Return a clone.

        Parameters
        ----------
        atom_map : :class:`dict`, optional
            If the clone should hold different :class:`.Atom`
            instances, then a :class:`dict` should be provided, which
            maps atoms in the current :class:`.FunctionalGroup` to the
            atoms which should be used in the clone. Only atoms which
            need to be remapped need to be present in the `atom_map`.

        Returns
        -------
        :class:`FunctionalGroup`
            A clone.

        Examples
        --------
        .. code-block:: python

            import stk

            c0 = stk.C(0)
            c1 = stk.C(1)
            bb = stk.BuildingBlock('NCCN', ['amine'])
            fg = bb.func_groups[0]
            a0, a1 = fg.atoms[:2]

            # fg_clone is a clone of fg, except that in all places
            # fg holds a0 fg_clone holds c0 and in all places where
            # fg holds a1, fg_clone holds c1.
            fg_clone = fg.clone({
                a0: c0
                a1: c1
            })

        """

        if atom_map is None:
            atom_map = {}

        return self.__class__(
            atoms=tuple(atom_map.get(a, a) for a in self.atoms),
            bonders=tuple(atom_map.get(a, a) for a in self.bonders),
            deleters=tuple(atom_map.get(a, a)for a in self.deleters),
            fg_type=self.fg_type
        )

    def get_atom_ids(self):
        """
        Yield the ids of :attr:`atoms` in order.

        Yields
        ------
        :class:`int`
            The id an :class:`.Atom` in :attr:`atoms`.

        """

        yield from (a.id for a in self.atoms)

    def get_bonder_ids(self):
        """
        Yield the ids of :attr:`bonders` in order.

        Yields
        ------
        :class:`int`
            The id of an :class:`.Atom` in :attr:`bonders`.

        """
        yield from (a.id for a in self.bonders)

    def get_deleter_ids(self):
        """
        Yield the ids of :attr:`deleters` in order.

        Yields
        -------
        :class:`int`
            The id of an :class:`.Atom` in :attr:`deleters`.

        """

        yield from (a.id for a in self.deleters)

    def __repr__(self):
        atoms = list(self.atoms)
        if len(atoms) == 1:
            atoms.append(None)
        atoms = ', '.join(
            repr(atom) if atom is not None else '' for atom in atoms
        )

        bonders = list(self.bonders)
        if len(bonders) == 1:
            bonders.append(None)
        bonders = ', '.join(
            repr(atom) if atom is not None else '' for atom in bonders
        )

        deleters = list(self.deleters)
        if len(deleters) == 1:
            deleters.append(None)
        deleters = ', '.join(
            repr(atom) if atom is not None else '' for atom in deleters
        )

        return (
            f'FunctionalGroup(\n'
            f'    atoms=( {atoms} ), \n'
            f'    bonders=( {bonders} ), \n'
            f'    deleters=( {deleters} ), \n'
            f'    fg_type={self.fg_type}\n'
            ')'
        )

    def __str__(self):
        atoms = list(self.atoms)
        if len(atoms) == 1:
            atoms.append('')
        atoms = ', '.join(str(atom) for atom in atoms)

        bonders = list(self.bonders)
        if len(bonders) == 1:
            bonders.append('')
        bonders = ', '.join(str(atom) for atom in bonders)

        deleters = list(self.deleters)
        if len(deleters) == 1:
            deleters.append('')
        deleters = ', '.join(str(atom) for atom in deleters)

        return (
            f'FunctionalGroup(\n'
            f'    atoms=( {atoms} ), \n'
            f'    bonders=( {bonders} ), \n'
            f'    deleters=( {deleters} ), \n'
            f'    fg_type={self.fg_type}\n'
            ')'
        )


_fg_types = (

    FGType(
        name='amine',
        func_group_smarts='[N]([H])[H]',
        bonder_smarts=['[$([N]([H])[H])]'],
        deleter_smarts=['[$([H][N][H])]']*2
    ),

    FGType(
        name='primary_amine',
        func_group_smarts=(
            '[N;$(NC);!$(NC=*);!$(NC#*);!$(NC(-[O,N,S]))]([H])[H]'
        ),
        bonder_smarts=['[$([N]([H])[H])]'],
        deleter_smarts=['[$([H][N][H])]']*2
    ),

    FGType(
        name='aldehyde',
        func_group_smarts='[C](=[O])[H]',
        bonder_smarts=['[$([C](=[O])[H])]'],
        deleter_smarts=['[$([O]=[C][H])]']
    ),

    FGType(
        name='carboxylic_acid',
        func_group_smarts='[C](=[O])[O][H]',
        bonder_smarts=['[$([C](=[O])[O][H])]'],
        deleter_smarts=[
            '[$([H][O][C](=[O]))]',
            '[$([O]([H])[C](=[O]))]'
        ]
    ),

    FGType(
        name='amide',
        func_group_smarts='[C](=[O])[N]([H])[H]',
        bonder_smarts=['[$([C](=[O])[N]([H])[H])]'],
        deleter_smarts=(
            ['[$([N]([H])([H])[C](=[O]))]'] +
            ['[$([H][N]([H])[C](=[O]))]']*2
        )
    ),

    FGType(
        name='thioacid',
        func_group_smarts='[C](=[O])[S][H]',
        bonder_smarts=['[$([C](=[O])[S][H])]'],
        deleter_smarts=[
            '[$([H][S][C](=[O]))]',
            '[$([S]([H])[C](=[O]))]'
        ]
    ),

    FGType(
        name='alcohol',
        func_group_smarts='[O][H]',
        bonder_smarts=['[$([O][H])]'],
        deleter_smarts=['[$([H][O])]']
    ),

    FGType(
        name='thiol',
        func_group_smarts="[S][H]",
        bonder_smarts=['[$([S][H])]'],
        deleter_smarts=['[$([H][S])]']
    ),

    FGType(
        name='bromine',
        func_group_smarts='*[Br]',
        bonder_smarts=['[$(*[Br])]'],
        deleter_smarts=['[$([Br]*)]']
    ),

    FGType(
        name='iodine',
        func_group_smarts='*[I]',
        bonder_smarts=['[$(*[I])]'],
        deleter_smarts=['[$([I]*)]']
    ),

    FGType(
        name='alkyne',
        func_group_smarts='[C]#[C][H]',
        bonder_smarts=['[$([C]([H])#[C])]'],
        deleter_smarts=['[$([H][C]#[C])]']
    ),

    FGType(
        name='terminal_alkene',
        func_group_smarts='[C]=[C]([H])[H]',
        bonder_smarts=['[$([C]=[C]([H])[H])]'],
        deleter_smarts=(
            ['[$([H][C]([H])=[C])]']*2 +
            ['[$([C](=[C])([H])[H])]']
        )
    ),

    FGType(
        name='boronic_acid',
        func_group_smarts='[B]([O][H])[O][H]',
        bonder_smarts=['[$([B]([O][H])[O][H])]'],
        deleter_smarts=(
            ['[$([O]([H])[B][O][H])]']*2 +
            ['[$([H][O][B][O][H])]']*2
        )
    ),

    # This amine functional group only deletes one of the
    # hydrogen atoms when a bond is formed.
    FGType(
        name='amine2',
        func_group_smarts='[N]([H])[H]',
        bonder_smarts=['[$([N]([H])[H])]'],
        deleter_smarts=['[$([H][N][H])]']
    ),

    FGType(
        name='secondary_amine',
        func_group_smarts='[H][N]([#6])[#6]',
        bonder_smarts=[
            '[$([N]([H])([#6])[#6])]'
        ],
        deleter_smarts=['[$([H][N]([#6])[#6])]']
    ),

    FGType(
        name='diol',
        func_group_smarts='[H][O][#6]~[#6][O][H]',
        bonder_smarts=['[$([O]([H])[#6]~[#6][O][H])]']*2,
        deleter_smarts=['[$([H][O][#6]~[#6][O][H])]']*2
    ),

    FGType(
        name='difluorene',
        func_group_smarts='[F][#6]~[#6][F]',
        bonder_smarts=['[$([#6]([F])~[#6][F])]']*2,
        deleter_smarts=['[$([F][#6]~[#6][F])]']*2
    ),

    FGType(
        name='dibromine',
        func_group_smarts='[Br][#6]~[#6][Br]',
        bonder_smarts=['[$([#6]([Br])~[#6][Br])]']*2,
        deleter_smarts=['[$([Br][#6]~[#6][Br])]']*2
    ),

    FGType(
        name='alkyne2',
        func_group_smarts='[C]#[C][H]',
        bonder_smarts=['[$([C]#[C][H])]'],
        deleter_smarts=['[$([H][C]#[C])]', '[$([C](#[C])[H])]']
    ),

    FGType(
        name='ring_amine',
        func_group_smarts='[N]([H])([H])[#6]~[#6]([H])~[#6R1]',
        bonder_smarts=[
            '[$([N]([H])([H])[#6]~[#6]([H])~[#6R1])]',
            '[$([#6]([H])(~[#6R1])~[#6][N]([H])[H])]',
        ],
        deleter_smarts=(
                ['[$([H][N]([H])[#6]~[#6]([H])~[#6R1])]']*2 +
                ['[$([H][#6](~[#6R1])~[#6][N]([H])[H])]']
        )
    ),

)

fg_types = {fg_type.name: fg_type for fg_type in _fg_types}



"""
Defines a class for each type of element.

Read the docstring of :class:`.Atom`.

"""


class Atom:
    """
    Represents an atom.

    A subclass is made for each element. The name of each subclass is
    the periodic table symbol of that element.

    Atoms of a particular element can be made with this
    class or with the subclass representing that element.

    Attributes
    ----------
    atomic_number : :class:`int`
        A class attribute. Specifies the atomic number.

    mass : :class:`float`
        A class attribute. Specifies the standard atomic weight.

    id : :class:`int`
        The id of the atom. Should be equal to its index in
        :attr:`.Molecule.atoms`.

    charge : :class:`int`
        The formal charge of the atom.

    Examples
    --------
    *Initialization.*

    Initialization of an :class:`.Atom` can happen in one of two ways.
    The atom can be initialized through the :class:`.Atom` class or
    through the class representing the element.

    .. code-block:: python

        import stk

        # h0 is an instance of the H class.
        h0 = stk.Atom(id=0, atomic_number=1)

        # h1 is also an instance of the H class.
        h1 = stk.H(id=1)

    When the class correspnding to the element is used directly, the
    ``atomic_number`` is not provided. Here are a few more examples.

    .. code-block:: python

        # Both he0 and he1 are instances of the He class.
        he0 = stk.Atom(id=2, atomic_number=2)
        he1 = stk.He(id=3)

        # Both c0 and c1 are instances of the
        # C class.
        c0 = stk.Atom(id=4, atomic_number=6)
        c1 = stk.C(id=5)

    *Adding additional atom attributes.*

    Each atom can be given additional attributes. For example

    .. code-block:: python

        c0.custom_attribute = 51
        c0.other_attr = 'something'

    We can initialize an :class:`.Atom` with additional, custom
    attributes directly

    .. code-block:: python

        h2 = stk.H(id=6, custom_attr1=123, other_attr2='thing')
        h2.custom_attr1  # Holds 123.
        h2.other_attr2  # Holds 'thing'.

    Providing the additional attributes to the initializer is
    functionally equivalent to assigning them to the object
    manually.

    *Printing*

    To print a brief summary of an atom you can run

    .. code-block:: python

        # Prints C(4).
        print(c0)

    To print a complete description of an atom, including additional
    attributes, you can run

    .. code-block:: python

        # Prints C(4, custom_attribute=51, other_attr='something').
        print(repr(c0))

    If private attributes were added to an atom, they will not be
    printed

    .. code-block:: python

        c0._attr_name = 12
        # Prints C(4, custom_attribute=51, other_attr='something').
        print(repr(c0))

    """

    # Maps each atomic number (int) to the relevant Atom subclass.
    _elements = {}

    def __init_subclass__(cls, **kwargs):
        # The init method for subclasses takes a slightly different
        # form.
        cls.__init__ = cls._subclass_init
        cls._elements[cls.atomic_number] = cls

    @staticmethod
    def _subclass_init(self, id, charge=0, **kwargs):
        """
        Initialize an atom of the element.

        Parameters
        ----------
        id : :class:`int`
            The id of the atom.

        charge : :class:`int`
            The formal charge.

        **kwargs : :class:`object`
            Additional attributes to be added to the atom.

        """

        Atom.__init__(self, id, self.atomic_number, charge, **kwargs)

    def __init__(self, id, atomic_number, charge=0, **kwargs):
        """
        Initialize an :class:`Atom`.

        Parameters
        ----------
        id : :class:`int`
            The id of the atom.

        atomic_number : :class:`int`
            The atomic number.

        charge : :class:`int`
            The formal charge.

        **kwargs : :class:`object`
            Additional attributes to be added to the atom.

        """

        self.id = id
        self.charge = charge
        self.__class__ = self._elements[atomic_number]
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def __repr__(self):
        charge = f', charge={self.charge}' if self.charge != 0 else ''
        mandatory = {'charge', 'id'}
        attrs = ', '.join(
            f'{attr}={val!r}' for attr, val in vars(self).items()
            if attr not in mandatory and not attr.startswith('_')
        )
        return (
            f'{self.__class__.__name__}'
            f'({self.id}{charge}{", " if attrs else ""}{attrs})'
        )

    def __str__(self):
        return f'{self.__class__.__name__}({self.id})'

    def clone(self):
        """
        Return a clone.

        Private attributes are not passed to the clone.

        Returns
        -------
        :class:`.Atom`
            The clone.

        """

        obj = self.__class__.__new__(self.__class__)
        for attr, val in vars(self).items():
            if not attr.startswith('_'):
                setattr(obj, attr, val)
        return obj


class H(Atom):
    atomic_number = 1
    mass = 1.008


class He(Atom):
    atomic_number = 2
    mass = 4.003


class Li(Atom):
    atomic_number = 3
    mass = 6.941


class Be(Atom):
    atomic_number = 4
    mass = 9.012


class B(Atom):
    atomic_number = 5
    mass = 10.811


class C(Atom):
    atomic_number = 6
    mass = 12.011


class N(Atom):
    atomic_number = 7
    mass = 14.007


class O(Atom):
    atomic_number = 8
    mass = 15.999


class F(Atom):
    atomic_number = 9
    mass = 18.998


class Ne(Atom):
    atomic_number = 10
    mass = 20.180


class Na(Atom):
    atomic_number = 11
    mass = 22.990


class Mg(Atom):
    atomic_number = 12
    mass = 24.305


class Al(Atom):
    atomic_number = 13
    mass = 26.982


class Si(Atom):
    atomic_number = 14
    mass = 28.086


class P(Atom):
    atomic_number = 15
    mass = 30.974


class S(Atom):
    atomic_number = 16
    mass = 32.066


class Cl(Atom):
    atomic_number = 17
    mass = 35.453


class Ar(Atom):
    atomic_number = 18
    mass = 39.948


class K(Atom):
    atomic_number = 19
    mass = 39.098


class Ca(Atom):
    atomic_number = 20
    mass = 40.078


class Sc(Atom):
    atomic_number = 21
    mass = 44.956


class Ti(Atom):
    atomic_number = 22
    mass = 47.867


class V(Atom):
    atomic_number = 23
    mass = 50.942


class Cr(Atom):
    atomic_number = 24
    mass = 51.996


class Mn(Atom):
    atomic_number = 25
    mass = 54.938


class Fe(Atom):
    atomic_number = 26
    mass = 55.845


class Co(Atom):
    atomic_number = 27
    mass = 58.933


class Ni(Atom):
    atomic_number = 28
    mass = 58.693


class Cu(Atom):
    atomic_number = 29
    mass = 63.546


class Zn(Atom):
    atomic_number = 30
    mass = 65.38


class Ga(Atom):
    atomic_number = 31
    mass = 69.723


class Ge(Atom):
    atomic_number = 32
    mass = 72.631


class As(Atom):
    atomic_number = 33
    mass = 74.922


class Se(Atom):
    atomic_number = 34
    mass = 78.971


class Br(Atom):
    atomic_number = 35
    mass = 79.904


class Kr(Atom):
    atomic_number = 36
    mass = 83.798


class Rb(Atom):
    atomic_number = 37
    mass = 85.468


class Sr(Atom):
    atomic_number = 38
    mass = 87.62


class Y(Atom):
    atomic_number = 39
    mass = 88.906


class Zr(Atom):
    atomic_number = 40
    mass = 91.224


class Nb(Atom):
    atomic_number = 41
    mass = 92.906


class Mo(Atom):
    atomic_number = 42
    mass = 95.95


class Tc(Atom):
    atomic_number = 43
    mass = 98.907


class Ru(Atom):
    atomic_number = 44
    mass = 101.07


class Rh(Atom):
    atomic_number = 45
    mass = 102.906


class Pd(Atom):
    atomic_number = 46
    mass = 106.42


class Ag(Atom):
    atomic_number = 47
    mass = 107.868


class Cd(Atom):
    atomic_number = 48
    mass = 112.414


class In(Atom):
    atomic_number = 49
    mass = 114.818


class Sn(Atom):
    atomic_number = 50
    mass = 118.711


class Sb(Atom):
    atomic_number = 51
    mass = 121.760


class Te(Atom):
    atomic_number = 52
    mass = 127.6


class I(Atom):
    atomic_number = 53
    mass = 126.904


class Xe(Atom):
    atomic_number = 54
    mass = 131.293


class Cs(Atom):
    atomic_number = 55
    mass = 132.905


class Ba(Atom):
    atomic_number = 56
    mass = 137.328


class La(Atom):
    atomic_number = 57
    mass = 138.905


class Ce(Atom):
    atomic_number = 58
    mass = 140.116


class Pr(Atom):
    atomic_number = 59
    mass = 140.908


class Nd(Atom):
    atomic_number = 60
    mass = 144.243


class Pm(Atom):
    atomic_number = 61
    mass = 144.913


class Sm(Atom):
    atomic_number = 62
    mass = 150.36


class Eu(Atom):
    atomic_number = 63
    mass = 151.964


class Gd(Atom):
    atomic_number = 64
    mass = 157.25


class Tb(Atom):
    atomic_number = 65
    mass = 158.925


class Dy(Atom):
    atomic_number = 66
    mass = 162.500


class Ho(Atom):
    atomic_number = 67
    mass = 164.930


class Er(Atom):
    atomic_number = 68
    mass = 167.259


class Tm(Atom):
    atomic_number = 69
    mass = 168.934


class Yb(Atom):
    atomic_number = 70
    mass = 173.055


class Lu(Atom):
    atomic_number = 71
    mass = 174.967


class Hf(Atom):
    atomic_number = 72
    mass = 178.49


class Ta(Atom):
    atomic_number = 73
    mass = 180.948


class W(Atom):
    atomic_number = 74
    mass = 183.84


class Re(Atom):
    atomic_number = 75
    mass = 186.207


class Os(Atom):
    atomic_number = 76
    mass = 190.23


class Ir(Atom):
    atomic_number = 77
    mass = 192.217


class Pt(Atom):
    atomic_number = 78
    mass = 195.085


class Au(Atom):
    atomic_number = 79
    mass = 196.967


class Hg(Atom):
    atomic_number = 80
    mass = 200.592


class Tl(Atom):
    atomic_number = 81
    mass = 204.383


class Pb(Atom):
    atomic_number = 82
    mass = 207.2


class Bi(Atom):
    atomic_number = 83
    mass = 208.980


class Po(Atom):
    atomic_number = 84
    mass = 208.982


class At(Atom):
    atomic_number = 85
    mass = 209.987


class Rn(Atom):
    atomic_number = 86
    mass = 222.018


class Fr(Atom):
    atomic_number = 87
    mass = 223.020


class Ra(Atom):
    atomic_number = 88
    mass = 226.025


class Ac(Atom):
    atomic_number = 89
    mass = 227.028


class Th(Atom):
    atomic_number = 90
    mass = 232.038


class Pa(Atom):
    atomic_number = 91
    mass = 231.036


class U(Atom):
    atomic_number = 92
    mass = 238.029


class Np(Atom):
    atomic_number = 93
    mass = 237.048


class Pu(Atom):
    atomic_number = 94
    mass = 244.064


class Am(Atom):
    atomic_number = 95
    mass = 243.061


class Cm(Atom):
    atomic_number = 96
    mass = 247.070


class Bk(Atom):
    atomic_number = 97
    mass = 247.070


class Cf(Atom):
    atomic_number = 98
    mass = 251.080


class Es(Atom):
    atomic_number = 99
    mass = 254


class Fm(Atom):
    atomic_number = 100
    mass = 257.095


class Md(Atom):
    atomic_number = 101
    mass = 258.1


class No(Atom):
    atomic_number = 102
    mass = 259.101


class Lr(Atom):
    atomic_number = 103
    mass = 262


class Rf(Atom):
    atomic_number = 104
    mass = 261


class Db(Atom):
    atomic_number = 105
    mass = 262


class Sg(Atom):
    atomic_number = 106
    mass = 266


class Bh(Atom):
    atomic_number = 107
    mass = 264


class Hs(Atom):
    atomic_number = 108
    mass = 269


class Mt(Atom):
    atomic_number = 109
    mass = 278


class Ds(Atom):
    atomic_number = 110
    mass = 281


class Rg(Atom):
    atomic_number = 111
    mass = 280


class Cn(Atom):
    atomic_number = 112
    mass = 285


class Nh(Atom):
    atomic_number = 113
    mass = 286


class Fl(Atom):
    atomic_number = 114
    mass = 289


class Mc(Atom):
    atomic_number = 115
    mass = 289


class Lv(Atom):
    atomic_number = 116
    mass = 293


class Ts(Atom):
    atomic_number = 117
    mass = 294


class Og(Atom):
    atomic_number = 118
    mass = 294



"""
Polymer
=======

#. :class:`.Linear`

"""

import logging
import numpy as np

from .topology_graph import TopologyGraph, VertexData, Vertex, EdgeData


logger = logging.getLogger(__name__)


class _LinearVertexData(VertexData):
    """
    Holds the data of a linear polymer vertex.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    flip : :class:`bool`
        If ``True`` any building block placed by the vertex will
        have its orientation along the chain flipped.

    """

    def __init__(self, x, y, z, flip):
        """
        Initialize a :class:`_LinearVertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        flip : :class:`bool`
            If ``True`` any building block placed by the vertex will
            have its orientation along the chain flipped.

        """

        self.flip = flip
        super().__init__(x, y, z)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone.flip = self.flip
        return clone

    def get_vertex(self):
        return _LinearVertex(self)


class _HeadVertexData(_LinearVertexData):
    def get_vertex(self):
        return _HeadVertex(self)


class _TailVertexData(_LinearVertexData):
    def get_vertex(self):
        return _TailVertex(self)


class _LinearVertex(Vertex):
    """
    Represents a vertex in the middle of a linear polymer chain.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        self._flip = data.flip
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            If ``True`` the :attr:`edges` attribute of the clone will
            be empty.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone._flip = self._flip
        return clone

    def place_building_block(self, building_block, vertices, edges):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        if len(building_block.func_groups) > 2:
            logger.warning(
                'You are placing a building block which has more than '
                'two functional groups along the backbone of '
                'a Linear topology. You can remove extra functional '
                'groups from the func_groups attribute to remove '
                'this message.'
            )

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids(fg_ids=(0, 1))
        )
        bonder_vector = next(
            building_block.get_bonder_direction_vectors(
                fg_ids=(0, 1)
            )
        )[-1]
        building_block.apply_rotation_between_vectors(
            start=bonder_vector,
            target=[-1 if self._flip else 1, 0, 0],
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Assign functional groups to edges.

        Each :class:`.FunctionalGroup` of the `building_block` needs
        to be associated with one of the :class:`.Edge` instances in
        :attr:`edges`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`dict`
            A mapping from the id of a functional group in
            `building_block` to the id of the edge in :attr:`edges` it
            is assigned to.

        """

        func_groups = building_block.func_groups
        fg1, fg2 = sorted(
            range(len(building_block.func_groups)),
            key=lambda fg_id: building_block.get_centroid(
                atom_ids=func_groups[fg_id].get_bonder_ids()
            )[0]
        )
        return {
            fg1: self._edge_ids[0],
            fg2: self._edge_ids[1]
        }

    def __str__(self):
        return (
            f'Vertex(id={self.id}, '
            f'position={self._position.tolist()}, '
            f'flip={self._flip})'
        )


class _TerminalVertex(_LinearVertex):
    """
    Represents a vertex at the end of a polymer chain.

    Do not instantiate this class directly, use :class:`.HeadVertex` or
    :class:`.TailVertex` instead.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def place_building_block(self, building_block, vertices, edges):
        """
        Place `building_block` on the :class:`.Vertex`.

        `building_block` is placed such that the centroid-centroid
        direction vector is always pointing away from the center of the
        chain, when `building_block` has only one
        :class:`.FunctionalGroup`.

        If the `building_block` has more than one
        :class:`.FunctionalGroup` then this method behaves in the same
        was as for :class:`.LinearVertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed on the :class:`.Vertex`.

        """

        if len(building_block.func_groups) != 1:
            return super().place_building_block(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids(fg_ids=(0, )),
        )
        centroid_vector = (
            building_block.get_centroid_centroid_direction_vector()
        )
        building_block.apply_rotation_between_vectors(
            start=centroid_vector,
            # _cap_direction is defined by a subclass.
            target=[self._cap_direction, 0, 0],
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Assign functional groups to edges.

        Each :class:`.FunctionalGroup` of the `building_block` needs
        to be associated with one of the :class:`.Edge` instances in
        :attr:`edges`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`dict`
            A mapping from the id of a functional group in
            `building_block` to the id of the edge in :attr:`edges` it
            is assigned to.

        Raises
        ------
        :class:`ValueError`
            If `building_block` does not have one or two functional
            groups.

        """

        if len(building_block.func_groups) == 2:
            func_groups = building_block.func_groups
            fgs = sorted(
                range(len(func_groups)),
                key=lambda fg_id: building_block.get_centroid(
                    atom_ids=func_groups[fg_id].get_bonder_ids()
                )[0]
            )
            fg_index = 0 if self._cap_direction == 1 else -1
            return {fgs[fg_index]: self._edge_ids[0]}

        elif len(building_block.func_groups) == 1:
            return {0: self._edge_ids[0]}

        else:
            raise ValueError(
                'The building block of a polymer '
                'must have 1 or 2 functional groups.'
            )


class _HeadVertex(_TerminalVertex):
    """
    Represents a vertex at the head of a polymer chain.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    # The direction to use if the building block placed on the
    # vertex only has 1 FunctionalGroup.
    _cap_direction = -1


class _TailVertex(_TerminalVertex):
    """
    Represents a vertex at the tail of a polymer chain.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    # The direction to use if the building block placed on the
    # vertex only has 1 FunctionalGroup.
    _cap_direction = 1


class Linear(TopologyGraph):
    """
    Represents a linear polymer topology graph.

    Attributes
    ----------
    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    Linear polymers require building blocks with functional groups

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CCC=O', ['aldehyde'])
        polymer = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.polymer.Linear('AB', 12)
        )

    However building blocks with a single functional group can
    also be provided as capping units

    .. code-block: python

        bb3 = stk.BuildingBlock('CCN', ['amine'])
        polymer = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=stk.polymer.Linear('ABABC', 1)
        )

    The repeating unit can also be specified through the indices of
    the building blocks

    .. code-block:: python

        # p1 and p2 are different ways to write the same thing.
        p1 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=stk.polymer.Linear('ACB', 1)
        )
        p2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3],
            topology_graph=stk.polymer.Linear((0, 2, 1), 1)
        )

    The `orientations` parameter allows the direction of each building
    block along to the chain to be flipped

    .. code-block:: python

        bb4 = stk.BuildingBlock('NCNCCN', ['amine'])

        p3 = stk.ConstructedMolecule(
            building_blocks=[bb2, bb4],
            topology_graph=stk.polymer.Linear(
                repeating_unit='AB',
                num_repeating_units=5,
                orientations=(1, 0.5)
            )
        )

    In the above example, ``bb1`` is guaranteed to be flipped,
    ``bb2`` has a 50 % chance of being flipped, each time it is placed
    on a node.

    Note that whether a building block will be flipped or not
    is decided during the initialization of :class:`.Linear`

    . code-block:: python

        # chain will always construct the same polymer.
        chain = stk.polymer.Linear(
            repeating_unit='AB',
            num_repeating_untis=5,
            orientations=(0.65, 0.45)
        )
        # p4 and p5 are guaranteed to be the same as they used the same
        # topology graph.
        p4 = stk.ConstructedMolecule([bb2, bb4], chain)
        p5 = stk.ConstructedMolecule([bb2, bb4], chain)

        # chain2 may lead to a different polymer than chain, despite
        # being initialized with the same parameters.
        chain2 = stk.polymer.Linear(
            repeating_unit='AB',
            num_repeating_untis=5,
            orientations=(0.65, 0.45)
        )

        # p6 and p7 are guaranteed to be the same because they used the
        # the same topology graph. However, they may be different to
        # p4 and p5.
        p6 = stk.ConstructedMolecule([bb2, bb4], chain2)
        p7 = stk.ConstructedMolecule([bb2, bb4], chain2)

    The `random_seed` parameter can be used to get reproducible results

    .. code-block:: python

        # p8 and p9 are guaranteed to be the same, because chain3 and
        # chain4 used the same random seeed.

        chain3 = stk.polymer.Linear(
            repeating_unit='AB',
            num_repeating_untis=5,
            orientations=(0.65, 0.45),
            random_seed=4
        )
        p8 = stk.ConstructedMolecule([bb2, bb4], chain3)

        chain4 = stk.polymer.Linear(
            repeating_unit='AB',
            num_repeating_untis=5,
            orientations=(0.65, 0.45),
            random_seed=4
        )
        p9 = stk.ConstructedMolecule([bb2, bb4], chain4)


    """

    def __init__(
        self,
        repeating_unit,
        num_repeating_units,
        orientations=None,
        random_seed=None,
        num_processes=1
    ):
        """
        Initialize a :class:`Linear` instance.

        Parameters
        ----------
        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            A string specifying the repeating unit of the polymer.
            For example, ``'AB'`` or ``'ABB'``. The first building
            block passed to `building_blocks` is ``'A'`` and so on.

            The repeating unit can also be specified by the indices of
            `building_blocks`, for example ``'ABB'`` can be
            written as ``(0, 1, 1)``.

        num_repeating_units : :class:`int`
            The number of repeating units which are used to make the
            polymer.

        orientations : :class:`tuple` of :class:`float`, optional
            For each character in the repeating unit, a value
            between ``0`` and ``1`` (both inclusive) must be given in
            a :class:`tuple`. It indicates the probability that each
            monomer will have its orientation along the chain flipped.
            If ``0`` then the monomer is guaranteed not to flip. If
            ``1`` it is guaranteed to flip. This allows the user to
            create head-to-head or head-to-tail chains, as well as
            chain with a preference for head-to-head or head-to-tail if
            a number between ``0`` and ``1`` is chosen. If ``None``
            then ``0`` is picked in all cases.

            It is also possible to supply an orientation for every
            vertex in the final topology graph. In this case, the
            length of `orientations` must be equal to
            ``len(repeating_unit)*num_repeating_units``.

        random_seed : :class:`int`, optional
            The random seed to use when choosing random orientations.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        Raises
        ------
        :class:`ValueError`
            If the length of `orientations` is not equal in length to
            `repeating_unit` or to the total number of vertices.

        """

        if orientations is None:
            orientations = tuple(
                0. for i in range(len(repeating_unit))
            )

        if len(orientations) == len(repeating_unit):
            orientations = orientations*num_repeating_units

        polymer_length = len(repeating_unit)*num_repeating_units
        if len(orientations) != polymer_length:
            raise ValueError(
                'The length of orientations must match either '
                'the length of repeating_unit or the '
                'total number of vertices.'
            )

        generator = np.random.RandomState(random_seed)

        # Keep these for __repr__.
        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        head, *body, tail = orientations
        choices = [True, False]
        vertex_data = [
            _HeadVertexData(
                x=0,
                y=0,
                z=0,
                flip=generator.choice(choices, p=[head, 1-head])
            )
        ]
        edge_data = []
        for i, p in enumerate(body, 1):
            flip = generator.choice(choices, p=[p, 1-p])
            v = _LinearVertexData(i, 0, 0, flip)
            vertex_data.append(v)
            edge_data.append(
                EdgeData(vertex_data[i-1], vertex_data[i])
            )

        vertex_data.append(
            _TailVertexData(
                x=len(vertex_data),
                y=0,
                z=0,
                flip=generator.choice(choices, p=[tail, 1-tail]))
        )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(int(v.flip) for v in vertex_data)

        edge_data.append(EdgeData(vertex_data[-2], vertex_data[-1]))

        super().__init__(
            vertex_data=tuple(vertex_data),
            edge_data=tuple(edge_data),
            construction_stages=(),
            num_processes=num_processes
        )

    @staticmethod
    def _normalize_repeating_unit(repeating_unit):
        if isinstance(repeating_unit, tuple):
            return repeating_unit

        base = ord('A')
        return tuple(ord(letter)-base for letter in repeating_unit)

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        """

        polymer = self._repeating_unit*self._num_repeating_units
        building_block_vertices = {}
        for bb_index, vertex in zip(polymer, self.vertices):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        return max(
            bb.get_maximum_diameter()
            for bb in mol.building_block_vertices
        )

    def __repr__(self):
        return (
            f'polymer.Linear({self._repeating_unit!r}, '
            f'{self._num_repeating_units!r}, '
            f'{self._orientations!r})'
        )



"""
Organic Cage
============

For usage examples see :class:`.Cage`.

#. :class:`.TwoPlusThree`
#. :class:`.FourPlusSix`
#. :class:`.FourPlusSix2`
#. :class:`.SixPlusNine`
#. :class:`.EightPlusTwelve`
#. :class:`.TwentyPlusThirty`
#. :class:`.TwoPlusFour`
#. :class:`.ThreePlusSix`
#. :class:`.FourPlusEight`
#. :class:`.FivePlusTen`
#. :class:`.SixPlusTwelve`
#. :class:`.EightPlusSixteen`
#. :class:`.TenPlusTwenty`
#. :class:`.SixPlusEight`
#. :class:`.OnePlusOne`
#. :class:`.TwoPlusTwo`
#. :class:`.FourPlusFour`
#. :class:`.TwelvePlusThirty`

"""


import numpy as np

from ..topology_graph import TopologyGraph, VertexData, Vertex
from ....utilities import vector_angle


class _CageVertexData(VertexData):
    """
    Holds the data of a cage vertex.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    aligner_edge : :class:`int`
        The edge which is used to align the :class:`.BuildingBlock`
        placed on the vertex. The first :class:`.FunctionalGroup`
        in :attr:`.BuildingBlock.func_groups` is rotated such that
        it lies exactly on this :class:`.Edge`. Must be between
        ``0`` and the number of edges the vertex is connected to.

    use_bonder_placement : :class:`bool`, optional
        If ``True``the position of the vertex will be updated such
        that it is in the middle of the neighboring bonder
        centroids, rather than in the middle of the neighboring
        vertices.

    """

    def __init__(self, x, y, z, use_bonder_placement=True):
        """
        Initialize a :class:`_CageVertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        use_bonder_placement : :class:`bool`, optional
            If ``True``the position of the vertex will be updated such
            that it is in the middle of the neighboring bonder
            centroids, rather than in the middle of the neighboring
            vertices.

        """

        self.aligner_edge = None
        self.use_bonder_placement = use_bonder_placement
        super().__init__(x, y, z)

    @classmethod
    def init_at_center(cls, *vertex_data):
        obj = super().init_at_center(*vertex_data)
        obj.aligner_edge = None
        obj.use_bonder_placement = True
        return obj

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`_CageVertexData`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone.aligner_edge = self.aligner_edge
        clone.use_bonder_placement = self.use_bonder_placement
        return clone

    def get_vertex(self):
        return _CageVertex(self)


class _CageVertex(Vertex):
    """
    Represents a vertex of a :class:`.Cage`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        # _neighbor_positions holds the bonder centroids of functional
        # groups on neighbor vertices connected to this vertex.
        self._neighbor_positions = []
        self._use_bonder_placement = data.use_bonder_placement
        # The edge which is used to align the :class:`.BuildingBlock`
        # placed on the vertex. The first :class:`.FunctionalGroup`
        # in :attr:`.BuildingBlock.func_groups` is rotated such that
        # it lies exactly on this :class:`.Edge`. Must be between
        # ``0`` and the number of edges the vertex is connected to.
        self._aligner_edge = data.aligner_edge
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone._aligner_edge = self._aligner_edge
        clone._use_bonder_placement = self._use_bonder_placement
        clone._neighbor_positions = list(self._neighbor_positions)
        return clone

    def get_aligner_edge(self):
        return self._aligner_edge

    def place_building_block(self, building_block, vertices, edges):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        if (
            self._use_bonder_placement
            and len(self._neighbor_positions) == len(self._edge_ids)
        ):
            self._update_position()

        if len(building_block.func_groups) == 2:
            return self._place_linear_building_block(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )
        return self._place_nonlinear_building_block(
            building_block=building_block,
            vertices=vertices,
            edges=edges
        )

    def _update_position(self):
        self._position = np.divide(
            np.sum(self._neighbor_positions, axis=0),
            len(self._neighbor_positions)
        )

    def _place_linear_building_block(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids()
        )
        fg_centroid = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        start = fg_centroid - self._position
        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        edge_coord = aligner_edge.get_position()
        connected_edges = tuple(edges[id_] for id_ in self._edge_ids)
        target = edge_coord - self._get_edge_centroid(
            centroid_edges=connected_edges,
            vertices=vertices
        )
        building_block.apply_rotation_between_vectors(
            start=start,
            target=target,
            origin=self._position
        )
        start = building_block.get_centroid_centroid_direction_vector()
        e0_coord = edges[self._edge_ids[0]].get_position()
        e1_coord = edges[self._edge_ids[1]].get_position()
        building_block.apply_rotation_to_minimize_angle(
            start=start,
            target=self._position,
            axis=e0_coord-e1_coord,
            origin=self._position,
        )
        return building_block.get_position_matrix()

    def _place_nonlinear_building_block(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids()
        )
        connected_edges = tuple(edges[id_] for id_ in self._edge_ids)
        edge_normal = self._get_edge_plane_normal(
            reference=self._get_edge_centroid(
                centroid_edges=connected_edges,
                vertices=vertices
            ),
            plane_edges=connected_edges,
            vertices=vertices
        )
        building_block.apply_rotation_between_vectors(
            start=building_block.get_bonder_plane_normal(),
            target=edge_normal,
            origin=self._position
        )
        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        start = fg_bonder_centroid - self._position
        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        edge_coord = aligner_edge.get_position()
        target = edge_coord - self._get_edge_centroid(
            centroid_edges=connected_edges,
            vertices=vertices
        )
        building_block.apply_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=edge_normal,
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Assign functional groups to edges.

        Each :class:`.FunctionalGroup` of the `building_block` needs
        to be associated with one of the :class:`.Edge` instances in
        :attr:`edges`.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`dict`
            A mapping from the id of a functional group in
            `building_block` to the id of the edge in :attr:`edges` it
            is assigned to.

        """

        if len(building_block.func_groups) == 2:
            return self._assign_func_groups_to_linear_edges(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )
        return self._assign_func_groups_to_nonlinear_edges(
                building_block=building_block,
                vertices=vertices,
                edges=edges
            )

    def after_assign_func_groups_to_edges(
        self,
        building_block,
        func_groups,
        vertices,
        edges
    ):
        """
        Perform operations after functional groups have been assigned.

        This method is always executed serially. It is often useful
        when data needs to be transferred between vertices, which
        have been processed independently, in parallel.

        Parameters
        ----------
        building_block : :class:`.Molecule`
            The building block molecule which is needs to have
            functional groups assigned to edges.

        func_groups : :class:`tuple` of :class:`.FunctionalGroup`
            The functional group clones added to the constructed
            molecule.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        None : :class:`NoneType`

        """

        bb_fgs = set(func_groups)
        for edge_id in self._edge_ids:
            for func_group in edges[edge_id].get_func_groups():
                if func_group not in bb_fgs:
                    continue

                bonder_position = self._get_molecule_centroid(
                    atom_ids=func_group.get_bonder_ids()
                )
                for vertex_id in edges[edge_id].get_vertex_ids():
                    if vertex_id == self.id:
                        continue

                    vertices[vertex_id]._neighbor_positions.append(
                        bonder_position
                    )

        return super().after_assign_func_groups_to_edges(
            building_block=building_block,
            func_groups=func_groups,
            vertices=vertices,
            edges=edges
        )

    def _assign_func_groups_to_linear_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return {
            fg_id: edge_id for fg_id, edge_id in enumerate(sorted(
                self._edge_ids,
                key=self._get_fg0_distance(building_block, edges)
            ))
        }

    def _get_fg0_distance(self, building_block, edges):
        fg_coord = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )

        def distance(edge_id):
            displacement = edges[edge_id].get_position() - fg_coord
            return np.linalg.norm(displacement)

        return distance

    def _assign_func_groups_to_nonlinear_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        # The idea is to order the functional groups in building_block
        # by their angle from func_groups[0] and the bonder centroid,
        #  going in the clockwise direction.
        #
        # The edges are also ordered by their angle from aligner_edge
        # and the edge centroid going in the clockwise direction.
        #
        # Once the fgs and edges are ordered, zip and assign them.

        fg0_coord = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        bonder_centroid = building_block.get_centroid(
            atom_ids=building_block.get_bonder_ids()
        )
        fg0_direction = fg0_coord-bonder_centroid
        axis = np.cross(
            fg0_direction,
            building_block.get_bonder_plane_normal()
        )
        func_groups = sorted(
            range(len(building_block.func_groups)),
            key=self._get_func_group_angle(
                building_block=building_block,
                fg0_direction=fg0_direction,
                bonder_centroid=bonder_centroid,
                axis=axis
            )
        )
        assignments = {}
        edge_ids = sorted(
            self._edge_ids,
            key=self._get_edge_angle(axis, vertices, edges)
        )
        for edge_id, fg_id in zip(edge_ids, func_groups):
            assignments[fg_id] = edge_id
        return assignments

    @staticmethod
    def _get_func_group_angle(
        building_block,
        fg0_direction,
        bonder_centroid,
        axis
    ):

        def angle(fg_id):
            func_group = building_block.func_groups[fg_id]
            coord = building_block.get_centroid(
                atom_ids=func_group.get_bonder_ids()
            )
            fg_direction = coord-bonder_centroid
            theta = vector_angle(fg0_direction, fg_direction)

            projection = fg_direction @ axis
            if theta > 0 and projection < 0:
                return 2*np.pi - theta
            return theta

        return angle

    def _get_edge_angle(self, axis, vertices, edges):
        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        aligner_edge_coord = aligner_edge.get_position()
        connected_edges = tuple(edges[id_] for id_ in self._edge_ids)
        edge_centroid = self._get_edge_centroid(
            centroid_edges=connected_edges,
            vertices=vertices
        )
        # This axis is used to figure out the clockwise direction.
        aligner_edge_direction = aligner_edge_coord - edge_centroid

        def angle(edge_id):
            coord = edges[edge_id].get_position()
            edge_direction = coord - edge_centroid
            theta = vector_angle(
                vector1=edge_direction,
                vector2=aligner_edge_direction
            )

            projection = edge_direction @ axis
            if theta > 0 and projection < 0:
                return 2*np.pi - theta
            return theta

        return angle

    def __str__(self):
        return (
            f'Vertex(id={self.id}, '
            f'position={self._position.tolist()}, '
            f'aligner_edge={self._aligner_edge})'
        )


class Cage(TopologyGraph):
    """
    Represents a cage topology graph.

    Cage topologies are added by creating a subclass which defines the
    :attr:`vertex_data` and :attr:`edge_data` class attributes.

    A :class:`Cage` subclass will add the attributes
    :attr:`num_windows` and :attr:`num_window_types` to each
    :class:`.ConstructedMolecule`.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    :class:`Cage` instances can be made without supplying
    additional arguments (using :class:`.FourPlusSix` as an example)

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CC(C=O)C=O', ['aldehyde'])
        cage1 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.cage.FourPlusSix()
        )

    Different structural isomers of cages can be made by using the
    `vertex_alignments` optional parameter

    .. code-block:: python

        tetrahedron = stk.cage.FourPlusSix(
            vertex_alignments={0: 1, 1: 1, 2: 2}
        )
        cage2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=tetrahedron
        )

    The parameter maps the :attr:`~.Vertex.id` of a vertex to a number
    between 0 (inclusive) and the number of edges the vertex is
    connected to (exclusive). So a vertex connected to three edges
    can be mapped to ``0``, ``1`` or ``2``.

    By changing which edge each vertex is aligned with, a different
    structural isomer of the cage can be formed.

    You can also build cages with multiple building blocks, but you
    have to assign each building block to a vertex with
    `building_block_vertices`.

    bb1 = stk.BuildingBlock('O=CC(C=O)C=O', ['aldehyde'])
    bb2 = stk.BuildingBlock('O=CC(Cl)(C=O)C=O', ['aldehyde'])
    bb3 = stk.BuildingBlock('NCCN', ['amine'])
    bb4 = stk.BuildingBlock('NCC(Cl)N', ['amine'])
    bb5 = stk.BuildingBlock('NCCCCN', ['amine'])

    tetrahedron = stk.cage.FourPlusSix()
    cage = stk.ConstructedMolecule(
        building_blocks=[bb1, bb2, bb3, bb4, bb5],
        topology_graph=tetrahedron,
        building_block_vertices={
            bb1: tetrahedron.vertices[:2],
            bb2: tetrahedron.vertices[2:4],
            bb3: tetrahedron.vertices[4:5],
            bb4: tetrahedron.vertices[5:6],
            bb5: tetrahedron.vertices[6:]
        }
    )

    """

    def __init_subclass__(cls, **kwargs):
        for i, vertex in enumerate(cls.vertex_data):
            vertex.id = i
        for i, edge in enumerate(cls.edge_data):
            edge.id = i
        return super().__init_subclass__(**kwargs)

    def __init__(self, vertex_alignments=None, num_processes=1):
        """
        Initialize a :class:`.Cage`.

        Parameters
        ----------
        vertex_alignments : :class:`dict`, optional
            A mapping from the :attr:`.Vertex.id` of a :class:`.Vertex`
            :attr:`vertices` to an :class:`.Edge` connected to it.
            The :class:`.Edge` is used to align the first
            :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
            placed on that vertex. Only vertices which need to have
            their default edge changed need to be present in the
            :class:`dict`. If ``None`` then the default edge is used
            for each vertex. Changing which :class:`.Edge` is used will
            mean that the topology graph represents different
            structural isomers. The edge is refered to by a number
            between ``0`` (inclusive) and the number of edges the
            vertex is connected to (exclusive).

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        """

        if vertex_alignments is None:
            vertex_alignments = {}

        vertex_data = {
            data: data.clone(True) for data in self.vertex_data
        }
        for vertex in vertex_data.values():
            vertex.aligner_edge = vertex_alignments.get(vertex.id, 0)
        edge_data = tuple(
            edge.clone(vertex_data)
            for edge in self.edge_data
        )
        vertex_types = sorted(
            {len(v.edges) for v in vertex_data},
            reverse=True
        )
        super().__init__(
            vertex_data=vertex_data.values(),
            edge_data=edge_data,
            construction_stages=tuple(
                lambda vertex, vertex_type=vt:
                    vertex.get_num_edges() == vertex_type
                for vt in vertex_types
            ),
            num_processes=num_processes
        )

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        Raises
        ------
        :class:`ValueError`
            If there is more than one building with a given number
            of functional groups.

        """

        bb_by_degree = {}
        for bb in building_blocks:
            num_fgs = len(bb.func_groups)
            if num_fgs in bb_by_degree:
                raise ValueError(
                    'If there are multiple building blocks with the '
                    'same number of functional groups, '
                    'building_block_vertices must be set explicitly.'
                )
            bb_by_degree[num_fgs] = bb

        building_block_vertices = {}
        for vertex in self.vertices:
            bb = bb_by_degree[vertex.get_num_edges()]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _prepare(self, mol):
        """
        Do preprocessing on `mol` before construction.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Order the building blocks by number of functional groups
        # so that building blocks with more functional groups are
        # always placed first.

        bb_verts = dict()
        bbs = sorted(
            mol.building_block_vertices,
            key=lambda bb: len(bb.func_groups),
            reverse=True
        )
        for bb in bbs:
            bb_verts[bb] = mol.building_block_vertices[bb]
        mol.building_block_vertices = bb_verts
        return super()._prepare(mol)

    def _clean_up(self, mol):
        mol.num_windows = self.num_windows
        mol.num_window_types = self.num_window_types
        return super()._clean_up(mol)

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        return max(
            bb.get_maximum_diameter()
            for bb in mol.building_block_vertices
        )

    def __repr__(self):
        vertex_alignments = ', '.join(
            f'{v.id}: {v.get_aligner_edge()}'
            for v in self.vertices
        )
        return (
            f'cage.{self.__class__.__name__}('
            f'vertex_alignments={{{vertex_alignments}}})'
        )
